"""Common utilities."""

import json
import logging
import re
from pathlib import Path

import numpy as np

from ewms_pilot.tasks.map import TaskMapping

LOGGER = logging.getLogger(__name__)


def all_task_errors_string(task_errors: list[BaseException]) -> str:
    """Make a string from the multiple task exceptions."""
    if not task_errors:
        return ""
    for exception in task_errors:
        LOGGER.error(exception)
    return (
        f"{len(task_errors)} TASK(S) FAILED: "
        f"{', '.join(repr(e) for e in task_errors)}"
    )


def dump_all_taskmaps(task_maps: list[TaskMapping]) -> None:
    """Dump all the task maps."""
    LOGGER.debug(
        json.dumps(
            [
                {
                    "start": tm.start_time,
                    "end": tm.end_time,
                    "runtime": tm.end_time - tm.start_time,
                    "done": tm.is_done,
                    "error": bool(tm.error),
                }
                for tm in task_maps
            ],
            indent=4,
        )
    )


def dump_task_runtime_stats(task_maps: list[TaskMapping]) -> None:
    """Dump stats about the given task maps."""
    LOGGER.info("Task runtime stats (successful tasks):")

    runtimes = [
        tm.end_time - tm.start_time for tm in task_maps if tm.is_done and not tm.error
    ]
    if not runtimes:
        LOGGER.info("no finished successful tasks")
        return

    data_np = np.array(runtimes)

    # calculate statistics
    stats_summary = {
        "Count": len(runtimes),
        "Mean": np.mean(data_np),
        "Median": np.median(data_np),
        "Variance": np.var(data_np),
        "Standard Deviation": np.std(data_np),
        "Min": np.min(data_np),
        "Max": np.max(data_np),
        "Range": np.ptp(data_np),
    }
    for key, value in stats_summary.items():
        LOGGER.info(f"({key.lower()}: {value:.2f})")

    def _to_range_string(_left: float, _right: float) -> str:
        return f"[{_left:.2f}, {_right:.2f})"

    # make bins and a terminal-friendly chart
    LOGGER.info("Runtimes distribution:")
    hist, bin_edges = np.histogram(data_np, bins="auto")
    no_datapoints_buffer: list[float] | None = None
    for i in range(len(hist)):
        # calc range bounds
        left = float(bin_edges[i])
        right = float(bin_edges[i + 1])

        # any data in range? if not, keep track of it so we don't log a ton of empty lines
        if not hist[i]:
            if no_datapoints_buffer:  # extend right bound
                no_datapoints_buffer[1] = right
            else:
                no_datapoints_buffer = [left, right]
            continue
        # now, we have datapoints. so, log whatever was built up
        if no_datapoints_buffer:
            bin_range = _to_range_string(
                no_datapoints_buffer[0], no_datapoints_buffer[1]
            )
            LOGGER.info(f"{bin_range:20} | [none]")
            no_datapoints_buffer = None

        # log it
        bar = "#" * hist[i]
        bin_range = _to_range_string(left, right)
        LOGGER.info(f"{bin_range:20} | {bar}")


def dump_tallies(task_maps: list[TaskMapping], dump_n_pending: bool = True) -> None:
    """Dump tallies about the given task maps."""
    string = ""
    if dump_n_pending:
        string += f"{TaskMapping.n_pending(task_maps)} Pending Tasks "
    string += (
        f"{TaskMapping.n_done(task_maps)} Finished Tasks "
        f"("
        f"{TaskMapping.n_successful(task_maps)} succeeded, "
        f"{TaskMapping.n_failed(task_maps)} failed"
        f")"
    )
    LOGGER.info(string)

    if _errors := [tm.error for tm in task_maps if tm.error]:
        LOGGER.error(all_task_errors_string(_errors))


class NoLogsInFileException(Exception):
    """Raised when there are no logs in a log file."""


class LogParser:
    """A parser for stderr logfiles."""

    # ex:
    # DEBUG   [U=59925,P=1]      SomeFunction()            Some log message
    # VERBOSE [U=613,P=42]       AnotherFunc()             Another log entry
    APPTAINER_LOG_PATTERN = re.compile(r"^[A-Z]+\s+\[U=\d+,P=\d+\]\s+")

    def __init__(self, log_fpath: Path):
        self.log_fpath = log_fpath

    def _get_nonblank_lines(self) -> list[str]:
        with open(self.log_fpath, "r", encoding="utf-8") as file:
            # no new-lines, no blank lines
            lines = [ln.rstrip("\n") for ln in file.readlines() if ln.strip()]
        if not lines:
            LOGGER.info("No lines in log file.")
            raise NoLogsInFileException()
        return lines

    def _get_last_non_apptainer_logline_index(self) -> int | None:
        lines = self._get_nonblank_lines()
        for i, line in enumerate(reversed(lines)):
            if not self.APPTAINER_LOG_PATTERN.match(line):
                return len(lines) - (i + 1)  # previous line's index
        return None  # aka it's all apptainer log lines

    def apptainer_extract_error(self) -> str:
        """Extracts the most relevant error message from a log file which includes Apptainer logs.

        Priority order:
        1. A Python traceback (if present).
        2. The first non-Apptainer error.
        3. The last Apptainer log entry, if nothing else is found.

        Returns:
            str: The most relevant error message found.
        """
        LOGGER.info(f"Extracting Apptainer logs from log file ({self.log_fpath})...")

        # prep
        try:
            lines = self._get_nonblank_lines()
        except NoLogsInFileException:
            LOGGER.info("No lines in log file.")
            return "<no stderr logs>"

        # Step 1: Locate the last error line before apptainer log line
        last_non_apptainer_index = self._get_last_non_apptainer_logline_index()

        # Step 2: Is there any actual good info here, or was this all apptainer logs?
        #
        # Example: "Child exited with exit status 255"
        # DEBUG   [U=613,P=1]        sylogBuiltin()                Running action command run
        # FATAL   [U=613,P=1]        StageTwo()                    exec /bin/bash failed: fork/exec /bin/bash: input/output error
        # DEBUG   [U=613,P=47]       startContainer()              stage 2 process reported an error, waiting status
        # DEBUG   [U=613,P=47]       CleanupContainer()            Cleanup container
        # DEBUG   [U=613,P=47]       umount()                      Umount /var/lib/apptainer/mnt/session/final
        # DEBUG   [U=613,P=47]       umount()                      Umount /var/lib/apptainer/mnt/session/rootfs
        # DEBUG   [U=613,P=47]       Master()                      Child exited with exit status 255
        # <EOF>
        if last_non_apptainer_index is None:  # still None b/c only saw apptainer logs
            LOGGER.info("Log file only contains apptainer logs--using the final one.")
            # return the very last line, parsed
            try:
                return "[Apptainer-Error] " + " ".join(
                    # Extract message
                    lines[-1].split(maxsplit=3)[3:]
                )
            except IndexError:
                LOGGER.warning(
                    f"failed to further parse error from apptainer-level log line "
                    f"(now, using whole line instead): {lines[-1]}"
                )
                return lines[-1]

        return self._extract_error(last_non_apptainer_index)

    def generic_extract_error(self) -> str:
        """Extracts the most relevant error message from a log file."""
        return self._extract_error(None)

    def _extract_error(self, last_line_index: int | None) -> str:
        """Extracts the most relevant error message from a log file.

        'last_line_index' controls how far to look into file (iow don't look at very end).
        """
        LOGGER.info(f"Extracting logs from log file ({self.log_fpath})...")

        # prep
        try:
            lines = self._get_nonblank_lines()
        except NoLogsInFileException:
            LOGGER.info("No lines in log file.")
            return "<no stderr logs>"
        if last_line_index is None:
            last_line_index = len(lines) - 1  # the actual last line

        # Step 1: Check for a Python traceback, then use that
        #
        # Example 1:
        # ...
        # Traceback (most recent call last):
        #   File "/usr/lib/python3.10/runpy.py", line 196, in _run_module_as_main
        #     return _run_code(code, main_globals, None,
        #   File "/usr/lib/python3.10/runpy.py", line 86, in _run_code
        #     exec(code, run_globals)
        # ...
        #   File "/usr/local/lib/python3.10/dist-packages/skymap_scanner/client/reco_icetray.py", line 151, in reco_pixel
        #     reco.setup_reco()
        #   File "/usr/local/lib/python3.10/dist-packages/skymap_scanner/recos/millipede_wilks.py", line 73, in setup_reco
        #     self.cascade_service = photonics_service.I3PhotoSplineService(
        # RuntimeError: Error reading table coefficients
        # <other lines skipped b/c 'last_line_index'>
        # <EOF>
        #
        # Example 2:
        # ...
        # Traceback (most recent call last):
        #   File "/usr/lib/python3.10/runpy.py", line 196, in _run_module_as_main
        #   File "/usr/lib/python3.10/runpy.py", line 86, in _run_code
        # ...
        #   File "<frozen importlib._bootstrap_external>", line 1016, in get_code
        #   File "<frozen importlib._bootstrap_external>", line 1073, in get_data
        # OSError: [Errno 107] Transport endpoint is not connected: '/usr/lib/python3/dist-packages/pandas/core/arrays/sparse/array.py'
        # <other lines skipped b/c 'last_line_index'>
        # <EOF>
        potential_python_traceback: list[str] = []
        for line in reversed(lines[: last_line_index + 1]):
            potential_python_traceback.insert(0, line)
            if line.startswith("Traceback"):  # Start of traceback found
                LOGGER.info("Logs contain a python traceback--using final traceback.")
                return "\n".join(potential_python_traceback)

        # ELSE: If no traceback, return last non-Apptainer error
        #
        # Example: "curl: (22) The requested URL returned error: 404"
        # ...
        # curl: (22) The requested URL returned error: 404
        # <other lines skipped b/c 'last_line_index'>
        # <EOF>
        LOGGER.info(
            f"Using {'last line' if last_line_index == len(lines) - 1 else f'line #{last_line_index}'}."
        )
        return lines[last_line_index]



"""Logic for running a subprocess."""

import asyncio
import dataclasses as dc
import json
import logging
import re
import shlex
import shutil
import subprocess
import sys
from pathlib import Path
from typing import TextIO

from .utils import LogParser
from ..config import (
    BIND_MOUNT_IN_CONTAINER_READONLY_DIRS,
    BIND_MOUNT_ON_PILOT_FORBIDDEN_DIRS,
    BIND_MOUNT_ON_PILOT_READONLY_DIRS,
    ENV,
    PILOT_DATA_DIR,
    PILOT_DATA_HUB_DIR_NAME,
)

LOGGER = logging.getLogger(__name__)

ENV_VAR_NAME_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")


# --------------------------------------------------------------------------------------


class ContainerSetupError(Exception):
    """Exception raised when a container pre-run actions fail."""

    def __init__(self, message: str, useful_identifier: str):
        """`useful_identifier` can be image name, task id, etc."""
        super().__init__(f"{message} for {useful_identifier}")


class ContainerRunError(Exception):
    """Raised when the container terminates in an error."""

    def __init__(
        self,
        alias: str,
        error_string: str,
        exit_code: int | None = None,
    ):
        exit_str = f" (exit code {exit_code})" if exit_code is not None else ""
        super().__init__(f"{alias} failed{exit_str}: {error_string}")


class InvalidBindMountError(Exception):
    """Raised when a bind mount is invalid."""


# --------------------------------------------------------------------------------------


@dc.dataclass
class ContainerBindMount:
    """A validated bind mount from host (on_pilot) to container (in_task_container)."""

    on_pilot: Path
    in_task_container: Path
    is_readonly: bool = False

    def __post_init__(self):
        for fpath in (self.on_pilot, self.in_task_container):
            if not fpath.is_absolute():
                raise InvalidBindMountError(
                    f"Bind mount path must be absolute: {fpath}"
                )

        if not self.on_pilot.is_dir():
            raise InvalidBindMountError(
                f"Bind mount does not exist on pilot: {self.on_pilot}"
            )

        # Disallow all access to forbidden host paths
        for forbidden in BIND_MOUNT_ON_PILOT_FORBIDDEN_DIRS:
            if self.on_pilot.resolve(strict=True).is_relative_to(forbidden):
                raise InvalidBindMountError(
                    f"Refusing to bind mount forbidden host path: {self.on_pilot}"
                )

        if not self.is_readonly:
            # Disallow writable mounts from read-only-only paths
            for readonly_only in BIND_MOUNT_ON_PILOT_READONLY_DIRS:
                if self.on_pilot.resolve(strict=True).is_relative_to(readonly_only):
                    raise InvalidBindMountError(
                        f"Refusing to bind mount writable host path: {self.on_pilot}"
                    )
            # Disallow writing into sensitive container paths
            for sensitive in BIND_MOUNT_IN_CONTAINER_READONLY_DIRS:
                if self.in_task_container.is_relative_to(sensitive):
                    raise InvalidBindMountError(
                        f"Refusing to mount writable directory to sensitive container path: {self.in_task_container}"
                    )


class DirectoryCatalog:
    """Handles the naming and mapping logic for a task's directories."""

    def __init__(self, name: str, include_task_io_directory: bool):
        """All directories are auto-created (task_io dir cannot already exist)."""
        self.name = name
        self._namebased_dir = PILOT_DATA_DIR / self.name

        def _mkdir(_fpath: Path, exist_ok=True) -> Path:
            _fpath.mkdir(parents=True, exist_ok=exist_ok)
            return _fpath

        # for inter-task/init storage: startup data, init container's output, etc.
        self.pilot_data_hub = ContainerBindMount(
            _mkdir(PILOT_DATA_DIR / PILOT_DATA_HUB_DIR_NAME),
            Path(f"/{PILOT_DATA_DIR.name}/{PILOT_DATA_HUB_DIR_NAME}"),
        )

        # for persisting stderr and stdout
        self.outputs_on_pilot = _mkdir(self._namebased_dir / "outputs")

        # for message-based task i/o
        if include_task_io_directory:
            self.task_io: ContainerBindMount | None = ContainerBindMount(
                _mkdir(self._namebased_dir / "task-io", exist_ok=False),
                Path(f"/{PILOT_DATA_DIR.name}/task-io"),
            )
        else:
            self.task_io = None

    def assemble_bind_mounts(
        self,
        include_external_directories: bool = False,
    ) -> list[ContainerBindMount]:
        """Get the docker bind mount string containing the wanted directories."""
        bind_mounts: list[ContainerBindMount] = [
            self.pilot_data_hub,
        ]

        if include_external_directories:
            bind_mounts.extend(
                ContainerBindMount(Path(d), Path(d), is_readonly=True)
                for d in ENV.EWMS_PILOT_EXTERNAL_DIRECTORIES.split(",")
                if d  # skip any blanks
            )

        if self.task_io:
            bind_mounts.append(self.task_io)

        targets = [m.in_task_container for m in bind_mounts]
        if len(set(targets)) != len(targets):
            raise ContainerSetupError(
                "Duplicate container mount targets detected",
                self.name,
            )

        return bind_mounts

    def rm_unique_dirs(self) -> None:
        """Remove all directories (on host) created for use only by this container."""
        shutil.rmtree(self._namebased_dir)  # rm -r


# --------------------------------------------------------------------------------------


def _dump_binary_file(fpath: Path, stream: TextIO, name: str) -> None:
    start_line = f"--- start: {name} ({stream.name}) "
    end_line = f"--- end: {name} ({stream.name}) "
    try:
        stream.write(start_line.ljust(60, "-") + "\n")
        stream.flush()
        with open(fpath, "rb") as file:
            while True:
                chunk = file.read(4096)
                if not chunk:
                    break
                stream.buffer.write(chunk)
        stream.write(end_line.ljust(60, "-") + "\n")
        stream.flush()
    except Exception as e:
        LOGGER.error(f"Error dumping container output ({stream.name}): {e}")


class ContainerRunner:
    """A utility class to run a container."""

    def __init__(
        self,
        image: str,
        args: str,
        timeout: int | None,
        env_json: str,
    ) -> None:
        self.args = args
        self.timeout = timeout
        self.image = self._prepull_image(image)

        if env := json.loads(env_json):
            LOGGER.debug(f"Validating env: {env}")
            # NOTE: there is additional validation before execution--the more validation, the better
            if not isinstance(env, dict) and not all(
                isinstance(k, str) and isinstance(v, (str | int))
                for k, v in env.items()
            ):
                raise ContainerSetupError(
                    "container's env must be a string-dictionary of strings or ints",
                    image,
                )
        else:
            env = {}
        self.env = env

    @staticmethod
    def _prepull_image(image: str) -> str:
        """Pull the image so it can be used in many tasks.

        Return the fully-qualified image name.
        """
        LOGGER.info(f"Pulling image: {image}")

        def _run(cmd: str):
            LOGGER.info(f"Running command: {cmd}")
            try:
                ret = subprocess.run(
                    cmd,
                    capture_output=True,  # redirect stdout & stderr
                    text=True,  # outputs are strings
                    check=True,  # raise if error
                    shell=True,
                )
                print(ret.stdout)
                print(ret.stderr, file=sys.stderr)
            except subprocess.CalledProcessError as e:
                print(e.stdout)
                print(e.stderr, file=sys.stderr)
                last_line = e.stderr.split("\n")[-1]
                raise ContainerSetupError(f"{str(e)} [{last_line}]", image)

        match ENV._EWMS_PILOT_CONTAINER_PLATFORM.lower():

            case "docker":
                if ENV.CI:  # optimization during testing, images are *loaded* manually
                    LOGGER.warning(
                        f"The pilot is running in a test environment, "
                        f"skipping 'docker pull {image}' (env var CI=True)"
                    )
                    return image
                _run(
                    #
                    # NOTE: validate & sanitize values HERE--this is the point of no return!
                    #       (making calls here makes it very clear what is checked)
                    #
                    f"docker pull "
                    f"{shlex.quote(image)}",
                )
                return image

            # NOTE: We are only are able to run unpacked directory format on condor.
            #       Otherwise, we get error: `code 255: FATAL:   container creation
            #       failed: image driver mount failure: image driver squashfuse_ll
            #       instance exited with error: squashfuse_ll exited: fuse: device
            #       not found, try 'modprobe fuse' first`
            #       See https://github.com/Observation-Management-Service/ewms-pilot/pull/86
            case "apptainer":
                if Path(image).exists() and Path(image).is_dir():
                    LOGGER.info("OK: Apptainer image is already in directory format")
                    return image
                elif ENV._EWMS_PILOT_APPTAINER_IMAGE_DIRECTORY_MUST_BE_PRESENT:
                    # not directory and image-conversions are disallowed
                    raise ContainerSetupError(
                        "Image 'not found in filesystem and/or "
                        "cannot convert to apptainer directory (sandbox) format",
                        image,
                    )
                # CONVERT THE IMAGE
                # assume non-specified image is docker -- https://apptainer.org/docs/user/latest/build_a_container.html#overview
                if "." not in image and "://" not in image:
                    # is not a blah.sif file (or other) and doesn't point to a registry
                    image = f"docker://{image}"
                # name it something that is recognizable -- and put it where there is enough space
                dir_image = (
                    f"{ENV._EWMS_PILOT_APPTAINER_BUILD_WORKDIR}/"
                    f"{image.replace('://', '_').replace('/', '_')}/"
                )
                # build (convert)
                _run(
                    #
                    # NOTE: validate & sanitize values HERE--this is the point of no return!
                    #       (making calls here makes it very clear what is checked)
                    #
                    # cd b/c want to *build* in a directory w/ enough space (intermediate files)
                    f"cd {ENV._EWMS_PILOT_APPTAINER_BUILD_WORKDIR} && "
                    f"apptainer "
                    f"{'--debug ' if ENV.EWMS_PILOT_CONTAINER_DEBUG else ''}"
                    f"build "
                    f"--fix-perms "
                    f"--sandbox {shlex.quote(dir_image)} "
                    f"{shlex.quote(image)}"
                )
                LOGGER.info(
                    f"Image has been converted to Apptainer directory format: {dir_image}"
                )
                return dir_image

            # ???
            case other:
                raise ValueError(
                    f"'_EWMS_PILOT_CONTAINER_PLATFORM' is not a supported value: {other}"
                )

    def _validate_env_var_name(self, name: str) -> str:
        if not ENV_VAR_NAME_RE.match(name):
            raise ContainerSetupError(
                f"Invalid environment variable name: {name}",
                self.image,
            )
        return name

    def _validate_env_var_value_to_str(self, value: str) -> str:
        if not isinstance(value, (str | int | Path)):
            raise ContainerSetupError(
                f"Invalid environment variable value (not str or int): '{value}'",
                self.image,
            )
        return str(value)

    async def run_container(
        self,
        logging_alias: str,  # what to call this container for logging and error-reporting
        stdoutfile: Path,
        stderrfile: Path,
        bind_mounts: list[ContainerBindMount],
        env_as_dict: dict,
        infile_arg_replacement: str = "",
        outfile_arg_replacement: str = "",
        datahub_arg_replacement: str = "",
    ) -> None:
        """Run the container and dump outputs."""
        dump_output = ENV.EWMS_PILOT_DUMP_TASK_OUTPUT

        # insert arg placeholder replacements
        # -> give an alternative for each token replacement b/c it'd be a shame if
        #    things broke this late in the game
        inst_args = self.args
        if infile_arg_replacement:
            for token in ["{{INFILE}}", "{{IN_FILE}}"]:
                inst_args = inst_args.replace(token, infile_arg_replacement)
        if outfile_arg_replacement:
            for token in ["{{OUTFILE}}", "{{OUT_FILE}}"]:
                inst_args = inst_args.replace(token, outfile_arg_replacement)
        if datahub_arg_replacement:
            for token in ["{{DATA_HUB}}", "{{DATAHUB}}"]:
                inst_args = inst_args.replace(token, datahub_arg_replacement)

        # assemble command
        # NOTE: don't add to bind_mounts (WYSIWYG); also avoid intermediate structures
        match ENV._EWMS_PILOT_CONTAINER_PLATFORM.lower():
            case "docker":
                cmd = (
                    #
                    # NOTE: validate & sanitize values HERE--this is the point of no return!
                    #       (making calls here makes it very clear what is checked)
                    #
                    f"docker run --rm "
                    # optional
                    f"{f'--shm-size={ENV._EWMS_PILOT_DOCKER_SHM_SIZE} ' if ENV._EWMS_PILOT_DOCKER_SHM_SIZE else ''}"
                    # bind mounts
                    f"{" ".join(
                        f"--mount type=bind,"
                        f"source={shlex.quote(str(m.on_pilot))},"
                        f"target={shlex.quote(str(m.in_task_container))}"
                        f"{',readonly' if m.is_readonly else ''}"
                        for m in bind_mounts
                    )} "  # <- space
                    # env vars
                    f"{" ".join(
                        f"--env "
                        f"{self._validate_env_var_name(n)}="
                        f"{shlex.quote(self._validate_env_var_value_to_str(v))}"
                        for n, v in sorted((self.env | env_as_dict).items())
                        # in case of key conflicts, choose the vals specific to this run
                    )} "  # <- space
                    # image + args
                    f"{shlex.quote(self.image)} "
                    f"{' '.join(shlex.quote(a) for a in shlex.split(inst_args))}"
                )
            case "apptainer":
                cmd = (
                    #
                    # NOTE: validate & sanitize values HERE--this is the point of no return!
                    #       (making calls here makes it very clear what is checked)
                    #
                    f"apptainer "
                    f"{'--debug ' if ENV.EWMS_PILOT_CONTAINER_DEBUG else ''}"
                    f"run "
                    # always add these flags
                    f"--containall "  # don't auto-mount anything
                    f"--no-eval "  # don't interpret CL args
                    # bind mounts
                    f"{" ".join(
                        f"--mount type=bind,"
                        f"source={shlex.quote(str(m.on_pilot))},"
                        f"target={shlex.quote(str(m.in_task_container))}"
                        f"{',readonly' if m.is_readonly else ''}"
                        for m in bind_mounts
                    )} "  # <- space
                    # env vars
                    f"{" ".join(
                        f"--env "
                        f"{self._validate_env_var_name(n)}="
                        f"{shlex.quote(self._validate_env_var_value_to_str(v))}"
                        for n, v in sorted((self.env | env_as_dict).items())
                        # in case of key conflicts, choose the vals specific to this run
                    )} "  # <- space
                    # image + args
                    f"{shlex.quote(self.image)} "
                    f"{' '.join(shlex.quote(a) for a in shlex.split(inst_args))}"
                )
            case other:
                raise ValueError(
                    f"'_EWMS_PILOT_CONTAINER_PLATFORM' is not a supported value: {other} ({logging_alias})"
                )
        LOGGER.info(f"Running {logging_alias} command: {cmd}")

        # run: call & check outputs
        try:
            with open(stdoutfile, "wb") as stdoutf, open(stderrfile, "wb") as stderrf:
                # await to start & prep coroutines
                proc = await asyncio.create_subprocess_shell(
                    cmd,
                    stdout=stdoutf,
                    stderr=stderrf,
                )
                # await to finish
                try:
                    await asyncio.wait_for(  # raises TimeoutError
                        proc.wait(),
                        timeout=self.timeout,
                    )
                except (TimeoutError, asyncio.exceptions.TimeoutError) as e:
                    # < 3.11 -> asyncio.exceptions.TimeoutError
                    raise ContainerRunError(
                        logging_alias,
                        f"[Timeout-Error] timed out after {self.timeout}s",
                    ) from e

            LOGGER.info(f"{logging_alias} return code: {proc.returncode}")

            # exception handling (immediately re-handled by 'except' below)
            if proc.returncode:
                log_parser = LogParser(stderrfile)
                raise ContainerRunError(
                    logging_alias,
                    (
                        log_parser.apptainer_extract_error()
                        if ENV._EWMS_PILOT_CONTAINER_PLATFORM.lower() == "apptainer"
                        else log_parser.generic_extract_error()
                    ),
                    exit_code=proc.returncode,
                )

        except Exception as e:
            LOGGER.error(f"{logging_alias} failed: {e}")  # log the time
            dump_output = True
            raise
        finally:
            if dump_output:
                _dump_binary_file(stdoutfile, sys.stdout, logging_alias)
                _dump_binary_file(stderrfile, sys.stderr, logging_alias)



"""Configuration constants."""

import dataclasses as dc
import enum
import logging
import os
from pathlib import Path

from wipac_dev_tools import from_environment_as_dataclass

LOGGER = logging.getLogger(__name__)


REFRESH_INTERVAL = 1  # sec -- the time between transitioning phases of the main loop


#
# Env var constants: set as constants & typecast
#


@dc.dataclass(frozen=True)
class EnvConfig:
    """For storing environment variables, typed."""

    #
    # REQUIRED
    #

    # task
    EWMS_PILOT_TASK_IMAGE: str = ""  # the image to run for each task
    EWMS_PILOT_TASK_ARGS: str = ""  # the args to run with the task container
    EWMS_PILOT_TASK_ENV_JSON: str = "{}"  # json dump of env variables + values to set

    # init
    EWMS_PILOT_INIT_IMAGE: str = ""  # the image to run once before processing any tasks
    EWMS_PILOT_INIT_ARGS: str = ""  # the args to run with the init container
    EWMS_PILOT_INIT_ENV_JSON: str = "{}"  # json dump of env variables + values to set

    # incoming queue
    EWMS_PILOT_QUEUE_INCOMING: str = ""  # name of the incoming queue
    EWMS_PILOT_QUEUE_INCOMING_AUTH_TOKEN: str = ""  # auth token for queue
    EWMS_PILOT_QUEUE_INCOMING_BROKER_TYPE: str = ""  # broker type: pulsar, rabbitmq...
    EWMS_PILOT_QUEUE_INCOMING_BROKER_ADDRESS: str = ""  # MQ broker URL to connect to

    # outgoing queue
    EWMS_PILOT_QUEUE_OUTGOING: str = ""  # name of the outgoing queue
    EWMS_PILOT_QUEUE_OUTGOING_AUTH_TOKEN: str = ""  # auth token for queue
    EWMS_PILOT_QUEUE_OUTGOING_BROKER_TYPE: str = ""  # broker type: pulsar, rabbitmq...
    EWMS_PILOT_QUEUE_OUTGOING_BROKER_ADDRESS: str = ""  # MQ broker URL to connect to

    #
    # OPTIONAL
    #

    EWMS_PILOT_DATA_DIR_PARENT_PATH_ON_HOST: str = ""

    # I/O to subprocess -- the file type (extension) of the input/output file from the pilot's task
    EWMS_PILOT_INFILE_EXT: str = ".in"  # ''
    EWMS_PILOT_OUTFILE_EXT: str = ".out"  # ''

    # incoming queue - settings
    EWMS_PILOT_PREFETCH: int = (
        1  # prefetch amount for incoming messages (off by default -- prefetch is an optimization)
    )
    EWMS_PILOT_TIMEOUT_QUEUE_WAIT_FOR_FIRST_MESSAGE: int | None = (
        None  # timeout (sec) for the first message to arrive at the pilot (defaults to incoming timeout value)
    )
    EWMS_PILOT_TIMEOUT_QUEUE_INCOMING: int = 1  # timeout (sec) for messages TO pilot

    # files
    EWMS_PILOT_EXTERNAL_DIRECTORIES: str = ""  # comma-delimited

    # logging -- only used when running via command line
    EWMS_PILOT_CL_LOG: str = "INFO"  # level for 1st-party loggers
    EWMS_PILOT_CL_LOG_THIRD_PARTY: str = "WARNING"  # level for 3rd-party loggers

    # chirp
    EWMS_PILOT_HTCHIRP: bool = False
    EWMS_PILOT_HTCHIRP_DEST: str = "JOB_ATTR"  # ["JOB_EVENT_LOG", "JOB_ATTR"]
    EWMS_PILOT_HTCHIRP_RATELIMIT_INTERVAL: float = 60.0

    # timing config -- tasks
    EWMS_PILOT_INIT_TIMEOUT: int | None = None  # timeout (sec) for the init command
    EWMS_PILOT_TASK_TIMEOUT: int | None = None  # timeout (sec) for each task

    # task handling logic
    EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR: bool = (
        True
        # whether to stop taking future tasks after a task fails;
        # ex: set to False if on known good compute node (testing cluster),
        #     set to True  if on unknown node (large homogeneous cluster)
    )
    EWMS_PILOT_OKAY_ERRORS: list[str] = dc.field(
        default_factory=list
        # applicable only when EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR=true;
        #    a list of names for the exceptions that will be ignored by the
        #    EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR setting
    )
    EWMS_PILOT_MAX_CONCURRENT_TASKS: int = 1  # max no. of tasks to process in parallel

    # misc settings
    EWMS_PILOT_KEEP_ALL_TASK_FILES: bool = False
    EWMS_PILOT_DUMP_TASK_OUTPUT: bool = (
        False  # dump each task's stderr to stderr and stdout to stdout
    )
    EWMS_PILOT_QUARANTINE_TIME: int = (
        0  # how long to sleep after error (useful for preventing blackhole scenarios on condor)
    )
    EWMS_PILOT_CONTAINER_DEBUG: bool = False

    # non-user set settings
    _EWMS_PILOT_CONTAINER_PLATFORM: str = "docker"
    _EWMS_PILOT_APPTAINER_BUILD_WORKDIR: str = "/var/tmp"
    _EWMS_PILOT_APPTAINER_IMAGE_DIRECTORY_MUST_BE_PRESENT: bool = True
    _EWMS_PILOT_DOCKER_SHM_SIZE: str | None = None  # this should be set to max allowed
    CI: bool = False  # github actions sets this to 'true'

    def __post_init__(self) -> None:
        """Do advanced validation."""

        # using the old env var?
        if timeout := os.getenv("EWMS_PILOT_SUBPROC_TIMEOUT"):
            LOGGER.warning(
                "Using 'EWMS_PILOT_SUBPROC_TIMEOUT'; 'EWMS_PILOT_TASK_TIMEOUT' is preferred."
            )
            if self.EWMS_PILOT_TASK_TIMEOUT is not None:
                LOGGER.warning(
                    "Ignoring 'EWMS_PILOT_SUBPROC_TIMEOUT' since 'EWMS_PILOT_TASK_TIMEOUT' was provided."
                )
            else:
                # b/c frozen
                object.__setattr__(self, "EWMS_PILOT_TASK_TIMEOUT", int(timeout))

        # must be positive
        if self.EWMS_PILOT_MAX_CONCURRENT_TASKS < 1:
            LOGGER.warning(
                f"Invalid value for 'EWMS_PILOT_MAX_CONCURRENT_TASKS' ({self.EWMS_PILOT_MAX_CONCURRENT_TASKS}),"
                " defaulting to '1'."
            )
            object.__setattr__(self, "EWMS_PILOT_CONCURRENT_TASKS", 1)  # b/c frozen

        # mutually exclusive
        if (
            self.EWMS_PILOT_QUARANTINE_TIME
            and not self.EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR
        ):
            raise RuntimeError(
                f"Cannot define 'EWMS_PILOT_QUARANTINE_TIME' while "
                f"'EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR' is "
                f"'{self.EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR}'"
            )

        # value must be one of these...
        if self.EWMS_PILOT_HTCHIRP_DEST not in ["JOB_EVENT_LOG", "JOB_ATTR"]:
            raise RuntimeError(
                f"Invalid EWMS_PILOT_HTCHIRP_DEST: {self.EWMS_PILOT_HTCHIRP_DEST}"
            )


ENV = from_environment_as_dataclass(EnvConfig)


# --------------------------------------------------------------------------------------


PILOT_DATA_DIR = Path(
    f"{ENV.EWMS_PILOT_DATA_DIR_PARENT_PATH_ON_HOST.rstrip('/')}/ewms-pilot-data"
)
PILOT_DATA_HUB_DIR_NAME = "data-hub"


class InTaskContainerEnvVarNames(enum.Enum):
    """The names of environment variables that are injected into the task container."""

    EWMS_TASK_DATA_HUB_DIR = enum.auto()
    EWMS_TASK_INFILE = enum.auto()
    EWMS_TASK_OUTFILE = enum.auto()


# --------------------------------------------------------------------------------------


# Paths in the container where writable mounts must be disallowed
BIND_MOUNT_IN_CONTAINER_READONLY_DIRS = [
    # Core system configuration and binaries
    Path("/etc"),  # Configuration files
    Path("/usr"),  # User-installed programs and data
    Path("/lib"),  # Shared libraries
    Path("/lib64"),  # 64-bit shared libraries
    Path("/bin"),  # Essential user binaries
    Path("/sbin"),  # System binaries
    # Sensitive user/system directories
    Path("/root"),  # Root user's home directory
    Path("/var"),  # Logs, caches, spool data
    # Virtual and kernel interfaces
    Path("/proc"),  # Kernel and process interface
    Path("/dev"),  # Device files
    Path("/sys"),  # Kernel sysfs interface
]

# Host paths that should never be mounted at all (even read-only)
BIND_MOUNT_ON_PILOT_FORBIDDEN_DIRS = [
    # Volatile and system-critical virtual filesystems
    Path("/proc"),  # Kernel and process interface
    Path("/dev"),  # Device files
    Path("/sys"),  # Kernel sysfs interface
    Path("/run"),  # Runtime state (e.g., PID files, sockets)
    # Boot and system startup
    Path("/boot"),  # Kernel, initramfs, bootloader
]

# Host paths that may be mounted read-only only (never writable)
BIND_MOUNT_ON_PILOT_READONLY_DIRS = [
    # Core system configuration and binaries
    Path("/etc"),  # Configuration files
    Path("/usr"),  # User-installed programs and data
    Path("/lib"),  # Shared libraries
    Path("/lib64"),  # 64-bit shared libraries
    Path("/bin"),  # Essential user binaries
    Path("/sbin"),  # System binaries
    # Sensitive user/system directories
    Path("/root"),  # Root user's home directory
    Path("/var"),  # Logs, caches, spool data
]


# --------------------------------------------------------------------------------------
